"""Fileman settings management."""
import logging
import os
from logging.config import dictConfig
from typing import Any, Dict, Iterable, Optional, Tuple, cast

import yaml
from pydantic import BaseSettings

try:  # pragma: no cover
    import sentry_sdk
    from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration
except ImportError:
    sentry_sdk = None
    SqlalchemyIntegration = None

_LOGGER = logging.getLogger(__name__)
ENVIRON_NAME = 'FILEMAN_CONFIG'
CONFIG_FILES = ('~/.fred/fileman.conf', '/etc/fred/fileman.conf')
# Mapping command line options to setting names.
OPTIONS_MAPPING = {
    '--connection': 'db_connection',
    '--max-workers': 'max_workers',
    '--poolclass': 'db_poolclass',
    '--port': 'grpc_port',
    '--storage': 'storage_root',
}


class FilemanSettings(BaseSettings):
    """Actual settings."""

    db_connection: str = 'sqlite:///:memory:'
    db_echo: bool = False
    db_poolclass: Optional[str] = None
    grpc_port: int = 50051
    logging: Optional[dict] = None
    max_workers: Optional[int] = None
    sentry: dict = {}
    storage_root: str = '/var/lib/fileman'


class Settings:
    """Settings manager - wraps messenger settings to take care of loading.."""

    def __init__(self) -> None:
        self._settings: Optional[FilemanSettings] = None

    def load(self, config_argv: Optional[str] = None) -> None:
        """Return setting object from the first config file found.

        Raises:
            FileNotFoundError: If the config file is provided, but the file doesn't exist.
            RuntimeError: If settings are already loaded.
        """
        if self._settings is not None:
            raise RuntimeError('Settings already loaded')
        config_file = os.environ.get(ENVIRON_NAME, config_argv)
        if not config_file:
            for name in CONFIG_FILES:
                full_name = os.path.expanduser(name)
                if os.path.isfile(full_name):
                    config_file = full_name
                    break
        if config_file:
            with open(config_file) as file:
                settings = yaml.safe_load(file)
        else:
            settings = {}
        self._settings = FilemanSettings(**settings)

        # Apply logging configuration
        if self['logging'] is not None:
            dictConfig(self['logging'])
        # Set up Sentry
        if self['sentry'] and self['sentry'].get('dsn'):
            if sentry_sdk is None:
                _LOGGER.warning('Sentry set up, but sentry-sdk not available.')
            else:
                sentry_sdk.init(dsn=self['sentry']['dsn'], integrations=[SqlalchemyIntegration()],
                                environment=self['sentry'].get('environment'), ca_certs=self['sentry'].get('ca_certs'))

    def __getitem__(self, key: str) -> Any:
        if self._settings is None:
            self.load()
        try:
            return getattr(self._settings, key)
        except AttributeError:
            raise KeyError(key)

    def __setitem__(self, key: str, value: Any) -> None:
        if self._settings is None:
            self.load()
        try:
            setattr(self._settings, key, value)
        except ValueError:
            raise KeyError(key)

    def items(self) -> Iterable[Tuple[str, Any]]:
        """Return an iterator of pairs of all settings."""
        if self._settings is None:
            self.load()
        return cast(FilemanSettings, self._settings).dict().items()

    def set_options(self, options: Dict[str, Any]) -> None:
        """Replace configuration values from command line options.

        Arguments:
            options: Command line argument parsed by docopt.
        """
        for option, setting in OPTIONS_MAPPING.items():
            value = options.get(option)
            if value:
                self[setting] = value


class SettingsProxy:
    """Proxy for settings instance.

    Attributes:
        settings: Actual settings
    """

    def __init__(self, settings: Settings):
        self.settings = settings

    def __getattr__(self, name: str) -> Any:
        return getattr(self.settings, name)

    def __getitem__(self, key: str) -> Any:
        return self.settings[key]

    def __setitem__(self, key: str, value: Any) -> Any:
        self.settings[key] = value


SETTINGS = SettingsProxy(Settings())
